import os
import cv2
import numpy as np
import sys
main_path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
if main_path not in sys.path:
    sys.path.append(main_path)
    sys.path.append(os.path.join(main_path, 'utils'))
from local_path import lasot_dir, tlp_dir, otb_dir, votlt19_dir, votlt18_dir
from moviepy.editor import *
import ffmpy

class p_config(object):

    name = 'bbox'
    save_img = True
    visualization = True

def get_groundtruth(Dataset, data_dir, video):
    if Dataset == "votlt" or Dataset == "votlt19":
        sequence_dir = data_dir + '/' + video + '/color/'
        gt_dir = data_dir + '/' + video + '/groundtruth.txt'
    elif Dataset == "otb":
        sequence_dir = data_dir + '/' + video + '/img/'
        gt_dir = data_dir + '/' + video + '/groundtruth_rect.txt'
    elif Dataset == "lasot":
        sequence_dir = data_dir + '/' + video + '/img/'
        gt_dir = data_dir + '/' + video + '/groundtruth.txt'
    elif Dataset == "tlp":
        sequence_dir = data_dir + '/' + video + '/img/'
        gt_dir = data_dir + '/' + video + '/groundtruth_rect.txt'
    try:
        groundtruth = np.loadtxt(gt_dir, delimiter=',')
    except:
        groundtruth = np.loadtxt(gt_dir)
    if Dataset == 'tlp':
        groundtruth = groundtruth[:, 1:5]

    return sequence_dir, groundtruth


def get_seq_list(Dataset, mode=None, classes=None, videos=None):
    if Dataset == "votlt18":
        data_dir = votlt18_dir
    elif Dataset == 'otb':
        data_dir = otb_dir
    elif Dataset == "votlt19":
        data_dir = votlt19_dir
    elif Dataset == "tlp":
        data_dir = tlp_dir
    elif Dataset == "lasot":
        data_dir = os.path.join(lasot_dir, classes)

    sequence_list = os.listdir(data_dir)
    sequence_list.sort()
    sequence_list = [title for title in sequence_list if not title.endswith("txt")]
    testing_set_dir = '../utils/testing_set.txt'
    testing_set = list(np.loadtxt(testing_set_dir, dtype=str))
    if mode == 'test' and Dataset == 'lasot':
        print('test data')
        sequence_list = [vid for vid in sequence_list if vid in testing_set]
    elif mode == 'train' and Dataset == 'lasot':
        print('train data')
        sequence_list = [vid for vid in sequence_list if vid not in testing_set]
    else:
        print("all data")
    if videos is not None:
        sequence_list = videos
    return sequence_list, data_dir


def run_seq_list(Dataset, p, sequence_list, data_dir):

    base_save_path = os.path.join('./plot_results', p.name, Dataset)

    for seq_id, video in enumerate(sequence_list):
        sequence_dir, groundtruth = get_groundtruth(Dataset, data_dir, video)

        image_list = os.listdir(sequence_dir)
        image_list.sort()
        image_list = [im for im in image_list if im.endswith("jpg") or im.endswith("jpeg")]

        for im_id in range(0, len(image_list)):
            imagefile = sequence_dir + image_list[im_id]
            image = cv2.imread(imagefile)
            gt = groundtruth[im_id]
            print("%d: " % seq_id + video + ": %d /" % im_id + "%d" % len(image_list))
            if not any(np.isnan(gt)):
                box = np.array([gt[0], gt[1], gt[0]+gt[2], gt[1]+gt[3]], dtype='int16').transpose()
                cv2.rectangle(image, (box[0], box[1]),
                              (box[2], box[3]), [0, 0, 255], 5)
            cv2.imshow(video, image)
            cv2.waitKey(0)

def save_seq_video(Dataset, p, sequence_list, data_dir):

    base_save_path = os.path.join('./plot_results', p.name, Dataset)

    for seq_id, video in enumerate(sequence_list):

        # Define the codec and create VideoWriter object
        # fourcc = cv2.cv.CV_FOURCC(*'DIVX')
        # out = cv2.VideoWriter('output.avi',fourcc, 20.0, (640,480))


        sequence_dir, groundtruth = get_groundtruth(Dataset, data_dir, video)

        image_list = os.listdir(sequence_dir)
        image_list.sort()
        image_list = [im for im in image_list if im.endswith("jpg") or im.endswith("jpeg")]
        image = cv2.imread(sequence_dir + image_list[0])
        height, width, layers = image.shape
        scale = 0.3
        size = (int(width*scale), int(height*scale))
        out = cv2.VideoWriter(video+'.avi', cv2.VideoWriter_fourcc(*'DIVX'), 15, size)
        for im_id in range(1, len(image_list), 5):
            imagefile = sequence_dir + image_list[im_id]
            image = cv2.imread(imagefile)
            gt = groundtruth[im_id]
            print("%d: " % seq_id + video + ": %d /" % im_id + "%d" % len(image_list))
            if not any(np.isnan(gt)):
                box = np.array([gt[0], gt[1], gt[0]+gt[2], gt[1]+gt[3]], dtype='int16').transpose()
                cv2.rectangle(image, (box[0], box[1]),
                              (box[2], box[3]), [0, 0, 255], 5)


            image = cv2.resize(image, (size))
            out.write(image)
            cv2.imshow(video, image)
            cv2.waitKey(1)
        out.release()
        ff = ffmpy.FFmpeg(
            inputs={video+".avi": None},
            outputs={video+".gif": None})
        ff.run()
if __name__ == '__main__':
    p = p_config()
    Dataset = 'votlt19'
    videos = ['deer']
    sequence_list, data_dir = get_seq_list(Dataset, videos=videos)
    save_seq_video(Dataset, p, sequence_list, data_dir)
    # clip = (VideoFileClip("PATH NAME").subclip((START TIME), (END TIME)).resize(ACCORDING TO THE USER WISH))
    # clip.write_gif("output.gif")
    # clip = (VideoFileClip("project.avi")).resize(0.3)
    # clip.write_gif("output.gif")


